from log import log
from rule import Rule
from symbols import Symbol, Symbols


class Edge(Rule):
    counter = 0

    def __str__(self):
        if self.stringified is None:
            self.stringified = Rule.__str__(self)
        return self.stringified

    def __repr__(self):
        return 'Edge(' + str(self) + ')'

    def __init__(self, predicted_from=None, predicted_by_rule=None,
                 extended_from=None, extended_using=None, tokens=None):
        Edge.counter += 1
        self.identificator = Edge.counter
        self.stringified = None

        self.predicted_from = predicted_from
        self.extended_from = extended_from
        if extended_from is not None:
            self._extend_from(extended_from, extended_using, tokens)
        elif predicted_by_rule is not None:
            self.rule = predicted_by_rule
            self.left = Symbol(predicted_by_rule.left)
            self.left.begin = 0
            self.left.length = 0
            Symbols.__init__(self, predicted_by_rule, position=0)
            if predicted_from is not None:
                self._predict_attributes()
                self.left.begin = predicted_from.left.end
        self.closed = self.position == len(self)
        self.current_symbol = None if self.closed else self[self.position]

    def _extend_from(self, extended_from, extended_using, tokens):
        self.predicted_from = extended_from.predicted_from
        self.rule = extended_from.rule
        self.left = Symbol(extended_from.left)
        Symbols.__init__(self, extended_from,
                         position=extended_from.position + 1)
        if extended_using is not None:
            self.left.length += extended_using.left.length
            self[extended_from.position].update(extended_using.left)
            self[extended_from.position].edge = extended_using
        elif tokens:  # terminal rule
            self.left.length += 1
            self[extended_from.position].update(
                tokens[extended_from.left.end])

    def copy_attributes(self, update=None):
        copy_from = self.position - 1
        if update:
            current = self[copy_from]
            for attr, value in sorted(update.items()):
                if value is None:
                    log.debug('null value   %s', attr)
                    continue  # asi by se nemělo kopírovat; nejspíš nemělo
                elif attr not in current or current[attr] is None:
                    log.debug('move         %s=%s → %s', attr, value, current)
                    current[attr] = value
                elif current[attr] != value:
                    log.warning('update       %s=%s → %s', attr, value,
                                current)
                    current[attr] = value

        for attr, indexes in sorted(self.rule.attr_matches.items()):
            if copy_from not in indexes:
                continue  # odsud se ani nemá kopírovat
            elif attr not in self[copy_from]:
                log.warning('no attr      %s %s', attr, self[copy_from])
                continue  # není co kopírovat

            value = self[copy_from][attr]
            if value is None:
                log.debug('null value   %s', attr)
                continue  # asi by se nemělo kopírovat; nejspíš nemělo

            if indexes[0] == -1:  # kopírování do hlavy
                log.debug('copy         %s=%s → head', attr, value)
                self.left[attr] = value

            # kopíruje se jen na členy vpravo od právě vloženého
            for index in indexes[indexes.index(copy_from)+1:]:
                log.debug('copy         %s=%s → %s', attr, value, self[index])
                self[index][attr] = value

    def _predict_attributes(self):
        symbol = self.predicted_from.current_symbol
        for attr, value in self.rule.left.items():
            if value is None and attr in symbol:
                self.left[attr] = symbol[attr]
                # log.debug('copy         %s=%s → head', attr, symbol[attr])

        for attr, match_on_indexes in sorted(self.rule.attr_matches.items()):
            if -1 in match_on_indexes and attr in self.left:
                value = self.left[attr]
                for index in match_on_indexes[1:]:
                    self[index][attr] = value
                    # log.debug('copy         head → %s=%s', attr, value)
