from collections import defaultdict
from itertools import chain, permutations

from symbols import Symbol, Symbols, escape


class Rule(Symbols):
    def __init__(self, rule=None, left=None, right=None):
        if rule is None:
            self.left = Symbol(left)
        else:
            if isinstance(rule, Rule):
                self.left = Symbol(rule.left)
                if right is None:
                    right = rule
            else:
                try:
                    left_side, right_side = rule.split(' → ')
                except ValueError:
                    raise ValueError('Multiple rules stuck together? ' + rule)
                self.left = Symbol(left_side)
                if right is None:
                    right = right_side
            if left:
                self.left.update(Symbol(left))
        Symbols.__init__(self, right)

        self.epsilon_rules = list(self._process_optional_symbols())
        self.attr_matches = self._process_matches()

    def __add__(self, value):
        return Rule(str(self) + str(value))

    def __eq__(self, other):
        # TODO: co něco jako
        # (INTER 1–2) → (OPT_COMMA 1–1) (NP g=I n=S c=1 1–2) (OPT_COMMA 2–2) •?
        # TODO: asi jen odsouvám problém dál, ani už nevím, kam :-D
        return str(self) == str(other)

    def __hash__(self):
        return hash(str(self))

    def __str__(self):
        return str(self.left) + ' → ' + Symbols.__str__(self)

    def __repr__(self):
        return 'Rule(' + escape(self) + ')'

    # TODO: byla to dobrá lekce, ale hodilo by se spíš se vrátit k __nonzero__,
    # které Python definoval pro Rule jako "is not None"; teď mi to rozděluje
    # pravidla od ε-pravidel, což vůbec nepotřebuju

    def _process_matches(self):
        attr_matches = defaultdict(list)  # → OrderedDict, abych nemusel řadit
        for index, symbol in enumerate([self.left] + list(self), -1):
            for attr, value in symbol.items():
                if value is None:
                    attr_matches[attr].append(index)

        for attr, match_on_indexes in attr_matches.items():
            if len(match_on_indexes) == 1:
                raise ValueError('No other symbol to match "%s" on in %s'
                                 % (attr, self))
        return attr_matches

    def _process_optional_symbols(self):
        for index, symbol in enumerate(self, 1):
            if symbol.optional:
                if 'phrase' not in symbol:
                    new_symbol = Symbol(symbol)
                    new_symbol.optional = False

                    # symbol['word'] = None
                    symbol.pop('word', None)
                    symbol.pop('lemma', None)
                    symbol.pop('tag', None)
                    symbol['phrase'] = self.left['phrase'] + str(index)

                    new_head = Symbol(symbol)
                    new_head.optional = False
                    yield Rule(left=new_head, right=[new_symbol])
                yield Rule(left=Symbol(phrase=symbol['phrase']), right=[])

    def insert_symbols(self, inserted_symbols, before=False, inside=False,
                       after=False, position=None, update_left=None):
        if isinstance(inserted_symbols, str):
            inserted_symbols = Symbols(inserted_symbols)

        left = Symbol(self.left)
        if isinstance(update_left, str):  # TODO: nebo radši jen replace_left?
            update_left = Symbol(update_left)
        if update_left:
            left.update(update_left)

        if position is not None:
            new_right = Symbols(self)
            for symbol in reversed(inserted_symbols):
                new_right.insert(position, symbol)
        else:
            new_right = Symbols(inserted_symbols if before else [])
            if inside:
                for symbol in self:
                    new_right.extend([symbol] + inserted_symbols)
                if not after:
                    for _ in range(len(inserted_symbols)):
                        new_right.pop()
            else:
                new_right.extend(self)
                if after:
                    new_right.extend(inserted_symbols)
        return Rule(left=left, right=new_right)

    def remove_attributes(self, removed_attributes, except_position=None,
                          repeat_self=False):
        if isinstance(removed_attributes, str):
            removed_attributes = Symbol(removed_attributes)
        new_right = Symbols()
        for index, symbol in enumerate(self):
            if index != except_position:
                symbol = Symbol(symbol)
                for attr in removed_attributes:
                    symbol.pop(attr, None)
            new_right.append(symbol)
        return Rule(left=self.left, right=new_right)

    # def insert_symbol_gradually(self, inserted_symbol, repeat_self=True):
    #     if isinstance(inserted_symbol, str):
    #         inserted_symbol = Symbol(inserted_symbol)
    #
    #     if repeat_self:
    #         yield self
    #
    #     for position in range(len(self) + 1):
    #         yield self.insert_symbol(inserted_symbol, position=position)

    # def permute(self):
    #     for permutation in permutations(self):
    #         yield Rule(left=self.left, right=permutation)


# def permute_rules(rules):
#     return chain.from_iterable(Rule(rule).permute() for rule in rules)


# def insert_gradually_into_rules(rules, inserted_symbol):
#     if isinstance(inserted_symbol, str):
#         inserted_symbol = Symbol(inserted_symbol)
#
#     return chain.from_iterable(
#         Rule(rule).insert_symbol_gradually(inserted_symbol) for rule in rules)
